from math import log, sqrt

class Edostep:
    def __init__(self, step, edo):
        self.step = step
        self.edo = edo
        if isinstance(self.step, int) and divisible_by(step * 1200, edo):
            self.cents = Cents(step * 1200 // edo)
        else:
            self.cents = Cents(step * 1200 / edo)

    def reduce(self):
        if isinstance(self.step, int) and isinstance(self.edo, int):
            r = Ratio(self.step, self.edo)
            self.step = r.numerator
            self.edo = r.denominator
            return self
        else:
            print('step and edo are not both ints')

    def __repr__(self):
        return '{}\\{}'.format(self.step, self.edo)

def divisible_by(n, d):
    return Ratio(n, d).denominator == 1

USE_SUBSCRIPT = False
subscript = dict.fromkeys([str(i) for i in range(10)]+['.','-'])
for i in range(10):
    subscript[str(i)] = chr(0x2080 + i)
subscript['-'] = '\u208b'
subscript['.'] = '.'
number = {v: k for k, v in subscript.items()}

def number_str_to_subscript(n):
    return ''.join([subscript[c] for c in str(n)])

def subscript_to_number_str(substr):
    return ''.join([number[c] for c in substr])

USE_UNICODE_ACCIDENTALS = False
unicode_accidental = {
    '\u266d': '\u266d',
    'b': '\u266d',
    ' ': ' ',
    '#': '\u266f',
    '\u266f': '\u266f'
}

class Pitch:
    def __init__(self, midi_key, cents=0):
        if type(midi_key) == str:
            # I know this is bad!
            self.midi_key = Pitch.from_name(midi_key).midi_key
            self.cents = Cents(cents)
        elif type(midi_key) == float:
            self.midi_key = int(midi_key)
            self.cents = Cents((midi_key % 1) * 100) + cents
            self.is_12_edo = False
        elif type(midi_key) == int:
            self.midi_key = midi_key
            self.cents = Cents(cents)
        elif type(midi_key) == Hertz:
            self.freq = midi_key
            midi_key = self.freq.to_midi_key()
            self.cents = Cents((midi_key % 1) * 100) + cents
            self.midi_key = int(midi_key)
        else:
            print("Exception")

        self.round_cents()
        self.set_name()
        self.set_freq() # This is unnecessary if self.freq is passed

    def midi(self):
        return self.midi_key

    def name(self):
        return self.spn()

    def spn(self):
        if USE_SUBSCRIPT:
            i = number_str_to_subscript(self.note_name_i)
        else:
            i = str(self.note_name_i)
        if USE_UNICODE_ACCIDENTALS:
            accidental = unicode_accidental[self.note_accidental].strip()
        else:
            accidental = self.note_accidental.strip()
        return self.note_name_base + accidental + i

    def round_cents(self):
        self.midi_key = self.midi_key + int((self.cents.value + 50) // 100)
        self.cents = Cents((self.cents + 50) % 100 - 50)

    def set_name(self):
        self.note_name_i = self.midi_key // 12 - 1
        sharp_accidentals_str = ' # #  # # # '
        sharp_names_str       = 'CCDDEFFGGAAB'
        i = int(self.midi_key % 12)
        self.note_name_base = sharp_names_str[i]
        self.note_accidental = sharp_accidentals_str[i]

    def from_name(name):
        sharp_names_str       = 'CCDDEFFGGAAB'
        midi_key = 0
        if name[0].isalpha():
            base_name = name[0]
            if name[1] in ['b', '#', chr(9837), chr(9838), chr(9839)]:
                accidental = name[1]
                if name[2].isdigit():
                    if name[2].isdecimal():
                        octave = int(name[2])
                    elif name[2] in number.keys():
                        octave = int(subscript_to_number_str(name[2]))
                else:
                    """ Exception """
                    pass
            elif name[1].isdigit():
                """ Negative numbers not yet included """
                accidental = None
                if name[1].isdecimal():
                    octave = int(name[1])
                elif name[1] in number.keys():
                    octave = int(subscript_to_number_str(name[1]))

            midi_key = sharp_names_str.find(base_name)
            if accidental in ['b', '\u266d']:
                midi_key -= 1
            if accidental in [None, chr(9840)]:
                """ Well, I guess do nothing """
                pass
            if accidental in ['#', '\u266f']:
                midi_key += 1
            octave += midi_key//12
            # C0 is 12
            midi_key += octave * 12 + 12
            return Pitch(midi_key)
            
        else:
            print(""" Exception """)
            pass

    def accidental(self):
        return self.accidental

    def set_freq(self):
        self.freq = Hertz(440 * 2**((self.midi_key-69 + self.cents.value/100.0) / 12))

    def get_freq(self):
        return self.freq

    def __eq__(self, other):
        return self.midi_key == other.midi_key

    def __gt__(self, other):
        return self.midi_key > other.midi_key

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other):
        return self.midi_key < other.midi_key

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, Interval12):
            return Pitch(self.midi_key + other.midi_interval * other.direction)
        elif isinstance(other, Interval94):
            return Pitch(self.midi_key, self.cents) + other.to_cents()
        elif isinstance(other, Cents):
            return Pitch(self.midi_key, self.cents + other.value)
        elif isinstance(other, Edostep):
            return Pitch(self.midi_key, self.cents + other.cents)

    def __sub__(self, other):
        if isinstance(other, Pitch):
            return Interval12(self, other)
        elif isinstance(other, Interval12):
            return self.__add__(-other)
        elif isinstance(other, Interval94):
            return Pitch(self.midi_key, self.cents) - other.to_cents()
        elif isinstance(other, Cents):
            return Pitch(self.midi_key, self.cents.value - other.value)
        elif isinstance(other, Edostep):
            return Pitch(self.midi_key, self.cents - other.cents)

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            p = 69 + 12*log2(self.get_freq() * other / 440.0)
            return Pitch(p)
        elif isinstance(other, Ratio):
            p = 69 + 12*log2(self.get_freq() * other.numerator / other.denominator / 440.0)
            return Pitch(p)

    def __truediv__(self, other):
        return self.__div__(other)

    def __div__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return self.__mul__(1 / other)
        elif isinstance(other, Ratio):
            return self.__mul__(other**-1)

    def __str__(self):
        if self.cents.value == 0:
            return self.spn()
        elif self.cents.value > 0:
            return self.spn() + ' +' + repr(self.cents)
        else:
            return self.spn() + ' ' + repr(self.cents)

    def __repr__(self):
        return self.__str__()

    def get_cents(self):
        pass

class Hertz(float):
    def to_midi_key(self):
        return 12 * log2(self.real / 440) + 69

    def __add__(self, other):
        if not isinstance(other, Cents):
            return Hertz(super().__add__(other))
        return Hertz(self.real * 2**(other.value / 1200.0))

    def __sub__(self, other):
        if not isinstance(other, Cents):
            return Hertz(super().__sub__(other))
        return Hertz(self.real / 2**(other.value / 1200.0))

    def __neg__(self):
        return Hertz(-self.real)

    def __mul__(self, other):
        return Hertz(self.real * other)

    def __rmul__(self, other):
        return Hertz(self.real * other)

    def __truediv__(self, other):
        return Hertz(self.real / other)

    def __rtruediv__(self, other):
        return Hertz(other / self.real)

    def __repr__(self):
        if self.real.is_integer():
            return str(int(self.real)) + ' Hz'
        elif (self.real * 10).is_integer():
            #return ('%.1f¢') % self.real
            return str(self.real) + ' Hz'
            # I think these are the same. Check Cents too
        else:
            return ('%.2f Hz') % self.real

NUMBER_OF_DECIMAL_PLACES_FOR_CENTS = 2

def to_cents(n):
    if isinstance(n, Cents):
        return n
    if isinstance(n, Interval94):
        return n.to_cents()
    if isinstance(n, int) or isinstance(n, float) or isinstance(n, Ratio):
        return 1200 * log2(n) * Cents()

class Cents:
    def __init__(self, value=None):
        if value == None:
            self.value = 1
        elif isinstance(value, self.__class__):
            self.value = value.value
        else:
            self.value = value
            
    def to_cents(self, cents):
        if isinstance(cents, self.__class__):
            return cents
        if isinstance(cents, int) or isinstance(cents, float):
            return Cents(cents)

    def __add__(self, other):
        return Cents(self.value + self.to_cents(other).value)
    def __sub__(self, other):
        if isinstance(other, int):
            return Cents(self.value - self.to_cents(other).value)
        elif isinstance(other, Pitch):
            return Pitch.__add__(other, self)
    def __neg__(self):
        return Cents(-self.value)
    def __mul__(self, other):
        return Cents(self.value * self.to_cents(other).value)
    def __rmul__(self, other):
        return self.__mul__(other)
    def __truediv__(self, other):
        return Cents(self.value / self.to_cents(other).value)
    def __mod__(self, other):
        return Cents(self.value % self.to_cents(other).value)
    def __floordiv__(self, other):
        return Cents(self.value // self.to_cents(other).value)
    
    def __repr__(self):
        if isinstance(self.value, int) or self.value.is_integer():
            return str(int(self.value)) + '¢'
        elif (self.value * 10).is_integer():
            if isinstance(self, float):
                return str(self.real) + '¢'
            else:
                return str(self.value) + '¢'
        else:
            return ('%.' + str(NUMBER_OF_DECIMAL_PLACES_FOR_CENTS) + 'f¢') % self.value

Intervals          = '122334T56677899AABBCDDEEF'
Interval_Qualities = 'PmMmMPTPmMmMPmMmMP#PmMmMP'

class PitchBend:
    def __init__(self, cents):
        # round is strange, round(.5) returns 0 but round(.51) returns 1

        # Here -100 is 0 but 100 is 8192. You need to review where 0 pb is!

        #if cents < 100
        # You need to fix this too
        self.value = round(((8192 * cents.value/100) + 8192)/2)

    def lsb(self):
        return self.value & 127

    def msb(self):
        return self.value >> 7

fifths = ['P1', 'P5', 'M2', 'M6', 'M3', 'M7', 'A4', 'm2', 'm6', 'm3', 'm7', 'P4']
fifths_41 = [
    'P1',
    'P5',
    'M2',
    'M6',
    'M3',
    'M7',
    'A4',
    '^m2',
    '^m6',
    '^m3',
    '^m7',
    '^4',
    '^1',
    '^5',
    '^M2',
    '^M6',
    '^M3',
    '^M7',
    '~5',
    '~2',
    '~6',
    '~3',
    '~7',
    '~4',
    'vm2',
    'vm6',
    'vm3',
    'vm7',
    'v4',
    'v8',
    'v5',
    'vM2',
    'vM6',
    'vM3',
    'vM7',
    'd5',
    'm2',
    'm6',
    'm3',
    'm7',
    'P4',
    #'P1',
]

fifths_94 = [ # negative piano
    'P1',          # P1       
    'P5',          # P5       
    'M2',          # M2       
    'M6',          # M6       
    'M3',          # M3       
    'M7',          # M7       
    'a4',          # Aug4     
    'pm2',         # ^m2      
    'pm6',         # ^m6      
    'pm3',         # ^m3      
    'pm7',         # ^m7      
    'p4',          # ^4       
    'p1',          # ^1       
    'p5',          # ^5       
    'pM2',         # ^M2      
    'pM6',         # ^M6      
    'pM3',         # ^M3      
    'pM7',         # ^M7      
    'pa4',         # ^Aug4    
    'ppm2',        # ^^m2     
    'ppm6',        # ^^m6     
    'ppm3',        # ^^m3     
    'ppm7',        # ^^m7     
    'llvd5',       # \\vdim5  
    'llvm2',       # \\vm2    
    'llvm6',       # \\vm6    
    'llvm3',       # \\vm3    
    'llvm7',       # \\vm7    
    'llv4',        # \\v4     
    'llv8',        # \\v8     
    'llv5',        # \\v5     
    'llvM2',       # \\vM2    
    'llvM6',       # \\vM6    
    'llvM3',       # \\vM3    
    'llvM7',       # \\vM7    
    'lld5',        # \\dim5   
    'llm2',        # \\m2     
    'llm6',        # \\m6     
    'llm3',        # \\m3     
    'llm7',        # \\m7     
    'll4',         # \\4      
    'll8',         # \\8      
    'll5',         # \\5      
    'llM2',        # \\M2     
    'llM6',        # \\M6     
    'llM3',        # \\M3     
    'llM7',        # \\M7     
    'lla4',        # \\Aug4   
    'ffm2',        # //m2     
    'ffm6',        # //m6     
    'ffm3',        # //m3     
    'ffm7',        # //m7     
    'ff4',         # //4      
    'ff1',         # //1      
    'ff5',         # //5      
    'ffM2',        # //M2     
    'ffM6',        # //M6     
    'ffM3',        # //M3     
    'ffM7',        # //M7     
    'ffa4',        # //Aug4   
    'ffpm2',       # //^m2    
    'ffpm6',       # //^m6    
    'ffpm3',       # //^m3    
    'ffpm7',       # //^m7    
    'ffp4',        # //^4     
    'ffp1',        # //^1     
    'ffp5',        # //^5     
    'ffpM2',       # //^M2    
    'ffpM6',       # //^M6    
    'ffpM3',       # //^M3    
    'ffpM7',       # //^M7    
    'ffpa4',       # //^Aug4  
    'vvM2',        # vvM2     
    'vvM6',        # vvM6     
    'vvM3',        # vvM3     
    'vvM7',        # vvM7     
    'vd5',         # vdim5    
    'vm2',         # vm2      
    'vm6',         # vm6      
    'vm3',         # vm3      
    'vm7',         # vm7      
    'v4',          # v4       
    'v8',          # v8       
    'v5',          # v5       
    'vM2',         # vM2      
    'vM6',         # vM6      
    'vM3',         # vM3      
    'vM7',         # vM7      
    'd5',          # dim5     
    'm2',          # m2       
    'm6',          # m6       
    'm3',          # m3       
    'm7',          # m7       
    'P4',          # P4       
    #'P1',         # P1       
]

edo94intervals = [
    'P1',              # P1      
    'ff1',             # //1     
    'p1',              # ^1      
    'ffp1',            # //^1    
    'llvm2',           # \\vm2   
    'vm2',             # vm2     
    'llm2',            # \\m2    
    'm2',              # m2      
    'ffm2',            # //m2    
    'pm2',             # ^m2     
    'ffpm2',           # //^m2   
    'ppm2',            # ^^m2    
    'vvM2',            # vvM2    
    'llvM2',           # \\vM2   
    'vM2',             # vM2     
    'llM2',            # \\M2    
    'M2',              # M2      
    'ffM2',            # //M2    
    'pM2',             # ^M2     
    'ffpM2',           # //^M2   
    'llvm3',           # \\vm3   
    'vm3',             # vm3     
    'llm3',            # \\m3    
    'm3',              # m3      
    'ffm3',            # //m3    
    'pm3',             # ^m3     
    'ffpm3',           # //^m3   
    'ppm3',            # ^^m3    
    'vvM3',            # vvM3    
    'llvM3',           # \\vM3   
    'vM3',             # vM3     
    'llM3',            # \\M3    
    'M3',              # M3      
    'ffM3',            # //M3    
    'pM3',             # ^M3     
    'ffpM3',           # //^M3   
    'llv4',            # \\v4    
    'v4',              # v4      
    'll4',             # \\4     
    'P4',              # P4      
    'ff4',             # //4     
    'p4',              # ^4      
    'ffp4',            # //^4    
    'llvd5',           # \\vd5   
    'vd5',             # vd5     
    'lld5',            # \\d5    
    'd5',              # d5      
    'lla4',            # \\a4    
    'a4',              # a4      
    'ffa4',            # //a4    
    'pa4',             # ^a4     
    'ffpa4',           # //^a4   
    'llv5',            # \\v5    
    'v5',              # v5      
    'll5',             # \\5     
    'P5',              # P5      
    'ff5',             # //5     
    'p5',              # ^5      
    'ffp5',            # //^5    
    'llvm6',           # \\vm6   
    'vm6',             # vm6     
    'llm6',            # \\m6    
    'm6',              # m6      
    'ffm6',            # //m6    
    'pm6',             # ^m6     
    'ffpm6',           # //^m6   
    'ppm6',            # ^^m6    
    'vvM6',            # vvM6    
    'llvM6',           # \\vM6   
    'vM6',             # vM6     
    'llM6',            # \\M6    
    'M6',              # M6      
    'ffM6',            # //M6    
    'pM6',             # ^M6     
    'ffpM6',           # //^M6   
    'llvm7',           # \\vm7   
    'vm7',             # vm7     
    'llm7',            # \\m7    
    'm7',              # m7      
    'ffm7',            # //m7    
    'pm7',             # ^m7     
    'ffpm7',           # //^m7   
    'ppm7',            # ^^m7    
    'vvM7',            # vvM7    
    'llvM7',           # \\vM7   
    'vM7',             # vM7     
    'llM7',            # \\M7    
    'M7',              # M7      
    'ffM7',            # //M7    
    'pM7',             # ^M7     
    'ffpM7',           # //^M7   
    'llv8',            # \\v8    
    'v8',              # v8      
    'll8',             # \\8     
    #'P8',             # P8      
]

class Interval94:
    # starting with octave reduced
    def __init__(self, interval_str_or_fifthspan):
        if isinstance(interval_str_or_fifthspan, str):
            # self.fifthspan = fifths_41.index(interval_str_or_fifthspan)
            self.fifthspan = fifths_94.index(interval_str_or_fifthspan)
            self.pynamestr = interval_str_or_fifthspan
        if isinstance(interval_str_or_fifthspan, int):
            self.fifthspan = interval_str_or_fifthspan
            self.pynamestr = fifths_94[self.fifthspan]

    def to_cents(self):
        return Cents(1200) * (edo94intervals.index(self.pynamestr)) / 94

    def __add__(self, other):
        if isinstance(other, Interval94):
            # return Interval94((self.fifthspan + other.fifthspan) % 41)
            return Interval94((self.fifthspan + other.fifthspan) % 94)

    def __sub__(self, other):
        if isinstance(other, Interval94):
            return Interval94((self.fifthspan - other.fifthspan) % 94)

    def __repr__(self):
        return fifths_94[self.fifthspan].replace('l', '\\').replace('f', '/').replace('p', '^').replace('a', 'Aug').replace('d', 'dim')

class Interval12: # Currently works for 12 only
    def __init__(self, a, b):
        # This is used for negating only
        self.a = a
        self.b = b

        self.midi_interval = abs(a.midi() - b.midi())
        self.direction = (int((a.midi() - b.midi()) >= 0) * 2) - 1
        if self.midi_interval == 6:
            self.interval_base = 'T'
        else:
            self.interval_base = str(int(Intervals[self.midi_interval], 16))
        self.quality = Interval_Qualities[self.midi_interval]

    def __str__(self):
        return (self.direction*'Ascending ' + -self.direction*'Descending ') + self.quality + self.interval_base

    def __repr__(self):
        return -self.direction*'-' + self.quality + self.interval_base + '//12'

    def __neg__(self):
        # Calculate Pitches and then give interval?
        return Interval12(self.b, self.a)


P1    = Interval94('P1')        # P1      
P5    = Interval94('P5')        # P5      
M2    = Interval94('M2')        # M2      
M6    = Interval94('M6')        # M6      
M3    = Interval94('M3')        # M3      
M7    = Interval94('M7')        # M7      
a4    = Interval94('a4')        # Aug4    
pm2   = Interval94('pm2')       # ^m2     
pm6   = Interval94('pm6')       # ^m6     
pm3   = Interval94('pm3')       # ^m3     
pm7   = Interval94('pm7')       # ^m7     
p4    = Interval94('p4')        # ^4      
p1    = Interval94('p1')        # ^1      
p5    = Interval94('p5')        # ^5      
pM2   = Interval94('pM2')       # ^M2     
pM6   = Interval94('pM6')       # ^M6     
pM3   = Interval94('pM3')       # ^M3     
pM7   = Interval94('pM7')       # ^M7     
pa4   = Interval94('pa4')       # ^Aug4   
ppm2  = Interval94('ppm2')      # ^^m2    
ppm6  = Interval94('ppm6')      # ^^m6    
ppm3  = Interval94('ppm3')      # ^^m3    
ppm7  = Interval94('ppm7')      # ^^m7    
llvd5 = Interval94('llvd5')     # \\vdim5 
llvm2 = Interval94('llvm2')     # \\vm2   
llvm6 = Interval94('llvm6')     # \\vm6   
llvm3 = Interval94('llvm3')     # \\vm3   
llvm7 = Interval94('llvm7')     # \\vm7   
llv4  = Interval94('llv4')      # \\v4    
llv8  = Interval94('llv8')      # \\v8    
llv5  = Interval94('llv5')      # \\v5    
llvM2 = Interval94('llvM2')     # \\vM2   
llvM6 = Interval94('llvM6')     # \\vM6   
llvM3 = Interval94('llvM3')     # \\vM3   
llvM7 = Interval94('llvM7')     # \\vM7   
lld5  = Interval94('lld5')      # \\dim5  
llm2  = Interval94('llm2')      # \\m2    
llm6  = Interval94('llm6')      # \\m6    
llm3  = Interval94('llm3')      # \\m3    
llm7  = Interval94('llm7')      # \\m7    
ll4   = Interval94('ll4')       # \\4     
ll8   = Interval94('ll8')       # \\8     
ll5   = Interval94('ll5')       # \\5     
llM2  = Interval94('llM2')      # \\M2    
llM6  = Interval94('llM6')      # \\M6    
llM3  = Interval94('llM3')      # \\M3    
llM7  = Interval94('llM7')      # \\M7    
lla4  = Interval94('lla4')      # \\Aug4  
ffm2  = Interval94('ffm2')      # //m2    
ffm6  = Interval94('ffm6')      # //m6    
ffm3  = Interval94('ffm3')      # //m3    
ffm7  = Interval94('ffm7')      # //m7    
ff4   = Interval94('ff4')       # //4     
ff1   = Interval94('ff1')       # //1     
ff5   = Interval94('ff5')       # //5     
ffM2  = Interval94('ffM2')      # //M2    
ffM6  = Interval94('ffM6')      # //M6    
ffM3  = Interval94('ffM3')      # //M3    
ffM7  = Interval94('ffM7')      # //M7    
ffa4  = Interval94('ffa4')      # //Aug4  
ffpm2 = Interval94('ffpm2')     # //^m2   
ffpm6 = Interval94('ffpm6')     # //^m6   
ffpm3 = Interval94('ffpm3')     # //^m3   
ffpm7 = Interval94('ffpm7')     # //^m7   
ffp4  = Interval94('ffp4')      # //^4    
ffp1  = Interval94('ffp1')      # //^1    
ffp5  = Interval94('ffp5')      # //^5    
ffpM2 = Interval94('ffpM2')     # //^M2   
ffpM6 = Interval94('ffpM6')     # //^M6   
ffpM3 = Interval94('ffpM3')     # //^M3   
ffpM7 = Interval94('ffpM7')     # //^M7   
ffpa4 = Interval94('ffpa4')     # //^Aug4 
vvM2  = Interval94('vvM2')      # vvM2    
vvM6  = Interval94('vvM6')      # vvM6    
vvM3  = Interval94('vvM3')      # vvM3    
vvM7  = Interval94('vvM7')      # vvM7    
vd5   = Interval94('vd5')       # vdim5   
vm2   = Interval94('vm2')       # vm2     
vm6   = Interval94('vm6')       # vm6     
vm3   = Interval94('vm3')       # vm3     
vm7   = Interval94('vm7')       # vm7     
v4    = Interval94('v4')        # v4      
v8    = Interval94('v8')        # v8      
v5    = Interval94('v5')        # v5      
vM2   = Interval94('vM2')       # vM2     
vM6   = Interval94('vM6')       # vM6     
vM3   = Interval94('vM3')       # vM3     
vM7   = Interval94('vM7')       # vM7     
d5    = Interval94('d5')        # dim5    
m2    = Interval94('m2')        # m2      
m6    = Interval94('m6')        # m6      
m3    = Interval94('m3')        # m3      
m7    = Interval94('m7')        # m7      
P4    = Interval94('P4')        # P4      
P1    = Interval94('P1')        # P1      

def factor(n):
    if not isinstance(n, int) or n == 0 or n == 1:
        return []
    for i in range(2, int(sqrt(n))+1):
        if n % i == 0:
            return [i] + factor(int(n/i))
    return [n]

def is_prime(n):
    return len(factor(n)) == 1

primes = []

for i in range(2, 65536):
    if is_prime(i):
        primes += [i]

def is_prime(n):
    return n in primes

import fractions
class Ratio:
    def __init__(self, numerator_or_str, denominator=1):
        if isinstance(numerator_or_str, Ratio):
            numerator, denominator = numerator_or_str.numerator, numerator_or_str.denominator
        if type(numerator_or_str) == float:
            f = fractions.Fraction(numerator_or_str).limit_denominator(1000)
            numerator, denominator = f.numerator, f.denominator
        elif type(numerator_or_str) == str:
            if numerator_or_str.count('/') == 1:
                numerator, denominator = [int(number_str) for number_str in numerator_or_str.split('/')]
            else:
                numerator = int(numerator_or_str)
        elif isinstance(numerator_or_str, int):
            numerator = numerator_or_str
        if denominator == 0:
            raise ZeroDivisionError
        self.is_negative = ((numerator < 0) ^ (denominator < 0)) and (numerator != 0)
        self.numerator = abs(numerator)
        self.denominator = abs(denominator)

        # Reduce
        for prime in set(factor(self.numerator)) & set(factor(self.denominator)):
            while prime in set(factor(self.numerator)) and prime in set(factor(self.denominator)):
                self.numerator //= prime
                self.denominator //= prime

        self.numerator_factors = factor(self.numerator)
        self.denominator_factors = factor(self.denominator)

    def __eq__(self, other):
        return self.is_negative == other.is_negative and self.numerator == other.numerator and self.denominator == other.denominator

    def __gt__(self, other):
        return -(int(self.is_negative)*2-1) * self.numerator / self.denominator > \
            -(int(other.is_negative)*2-1) * other.numerator / other.denominator

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other):
        return -(int(self.is_negative)*2-1) * self.numerator / self.denominator < \
            -(int(other.is_negative)*2-1) * other.numerator / other.denominator

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, int):
            return Ratio(-(int(self.is_negative)*2-1) * self.numerator + other * self.denominator, self.denominator)
        elif isinstance(other, Ratio):
            return Ratio(-(int(self.is_negative)*2-1) * self.numerator * other.denominator + -(int(other.is_negative)*2-1) * other.numerator * self.denominator, self.denominator * other.denominator)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        if self.numerator != 0:
            self.is_negative = not self.is_negative
        return self

    def __mul__(self, other):
        if isinstance(other, int):
            return Ratio(self.numerator * other, self.denominator)
        elif isinstance(other, Ratio):
            return Ratio(self.numerator * other.numerator, self.denominator * other.denominator)

    def __truediv__(self, other):
        if isinstance(other, int):
            return Ratio(self.numerator, self.denominator * other)
        elif isinstance(other, Ratio):
            return Ratio(self.numerator * other.denominator, self.denominator * other.numerator)

    def __floordiv__(self, other):
        if isinstance(other, int):
            return Cents(1200) * float(self)/other

    def __pow__(self, other):
        if isinstance(other, int):
            r = Ratio(self.numerator, self.denominator)
            was_negative = r.is_negative
            if r.is_negative:
                r.is_negative = other % 2 == 1
            if other > 1:
                for i in range(other - 1):
                    r.numerator *= r.numerator
                    r.denominator *= r.denominator
            elif other == 1:
                pass
            elif other == 0:
                r.numerator = 1
                r.denominator = 1
            elif other < 0:
                old_numerator = r.numerator
                r.numerator = r.denominator
                r.denominator = old_numerator
                if other < -1:
                    for i in range(-other - 1):
                        r.numerator *= r.numerator
                        r.denominator *= r.denominator
            return r

    def __float__(self):
        return self.numerator / self.denominator

    def monzo(self):
        monzo_list = []

        if len(self.numerator_factors) != 0:
            numerator_monzo_len = primes.index(self.numerator_factors[-1])
        else:
            numerator_monzo_len = 0
        if len(self.denominator_factors) != 0:
            denominator_monzo_len = primes.index(self.denominator_factors[-1])
        else:
            denominator_monzo_len = 0

        if numerator_monzo_len > denominator_monzo_len:
            monzo_len = numerator_monzo_len
        else:
            monzo_len = denominator_monzo_len

        for i in range(monzo_len + 1):
            monzo_list.append(self.numerator_factors.count(primes[i]) - self.denominator_factors.count(primes[i]))
        return Monzo(monzo_list)

    def __repr__(self):
        return '{}{}/{}'.format('-'*self.is_negative, self.numerator, self.denominator)

    def factor_str(self):
        if len(self.numerator_factors) > 0:
            s = str(self.numerator_factors[0])
        else:
            s = '1'
        if len(self.numerator_factors) > 1:
            for n in self.numerator_factors[1:]:
                s += ' * ' + str(n)
        if len(self.denominator_factors) > 0:
            s += ' / ' + str(self.denominator_factors[0])
        if len(self.denominator_factors) > 1:
            for n in self.denominator_factors[1:]:
                s += ' / ' + str(n)
        return s

class Monzo:
    def __init__(self, monzo_list):
        self.monzo_list = monzo_list

    def ratio(self):
        numerator = 1
        denominator = 1
        for i in range(len(self)):
            if self.monzo_list[i] > 0:
                numerator *= primes[i]**self.monzo_list[i]
            elif self.monzo_list[i] < 0:
                denominator *= primes[i]**-self.monzo_list[i]
        return Ratio(numerator, denominator)

    def to_list(self):
        return self.monzo_list

    def __getitem__(self, index):
        return self.monzo_list[index]

    def prime(self, prime_number):
        if prime_number in primes:
            index = primes.index(prime_number)
            if index < len(self):
                return self.monzo_list[primes.index(prime_number)]
            else:
                return 0
        else:
            print('error')

    def __len__(self):
        return len(self.monzo_list)

    def __repr__(self):
        # https://www.compart.com/en/unicode/U+27E9
        return '[{}⟩'.format(str(self.monzo_list)[1:-1])

def log2(x):
    return log(x) / log(2)
